# AC3D
# Copyright (c) 2016, Inivis Limited. All rights reserved.

# TCL stuff to handle editing of palette entries


# popup menu bound to each palette button

menu .palette_menu -tearoff 0
.palette_menu add command -label "Edit material..." -command {create_palette_window  $current_palette_index}
.palette_menu add command -label "Select surfaces with this material" -command {ac3d select_surfaces_by_material $current_palette_index }
.palette_menu add command -label "Append new material" -command "new_palette_entry"



proc popup_palette_menu { x y entry colindex} {
global current_palette_edit select_mode current_palette_index

#puts "popup_palette_menu $entry $colindex"

	if { ($select_mode == "surface") || ($select_mode == "vertex") } {
		.palette_menu entryconfigure 1 -state normal
	} else {
		.palette_menu entryconfigure 1 -state disabled
	}

    set current_palette_edit $entry
	set current_palette_index $colindex
    tk_popup .palette_menu $x $y

}


proc add_palette_button { colrgb colindex } {
#global .palette_frame.canvas
global tcl_platform ac_platform
global cp ac_platform
global UI

    set parent $UI(palette_inner_frame)

set newname [ format "$parent.%04d" $colindex ]
#puts "ADD_PALETTE BUTTON $colrgb $colindex"

if {$tcl_platform(platform) == "windows"} {
        ::tk::button $newname -text $colindex -background $colrgb \
            -command "ac3d set_col $colindex" -borderwidth 0 -padx 1 -pady 0 \
            -activebackground $colrgb -highlightthickness 0\
            -bitmap "" -default disabled
        $newname configure -width 3 -height 1

} else {

#	if { $ac_platform == "Mac" } {
            ::tk::label $newname -background $colrgb -width 3 -height 1 -borderwidth 0 -text $colindex -relief raised
            bind $newname <Button-1> "$newname configure -relief sunken"
		bind $newname <ButtonRelease-1> "ac3d set_col $colindex; $newname configure -relief raised"
#	} else {
#
#            ::tk::button $newname -text $colindex -background $colrgb \
#                -command "ac3d set_col $colindex" -borderwidth 1 -padx 0 -pady 0 \
#                -activebackground $colrgb -highlightthickness 0\
#                -bitmap ""
#            $newname configure -width 3 -height 1
#        }

}


pack $newname -side left -in $parent -padx { 0 1} -expand 1

    bind $newname <3> "popup_palette_menu %X %Y %W $colindex"
    bind $newname <2> "popup_palette_menu %X %Y %W $colindex"


    # on the mac, make control+mouse button do the same as button3
if { $ac_platform == "Mac" } {
	bind $newname <Control-ButtonPress-1> {event generate %W <ButtonPress-3> \
											   -x %x -y %y -rootx %X -rooty %Y -button 3 -time %t}
}

    add_balloon $newname "Set all selected surfaces to this col\n(right click for a popup menu)"

#mac needs this update to make the palette buttons show when AC3D starts?
#update idletasks
}





proc remove_palette_button { id } {
# called from inside AC3D when a material is removed
# umm, not yet implemented...

	set p [get_palette_button_path id]

	# need to remove the button and renumber the others... yuk
}









proc new_palette_entry { } {
global current_palette_edit UI

    # extract the number of the palette entry from the end of the buttns name
    # there must be an easier way of doing this...
    set buttonid [ string range $current_palette_edit [expr [ string length $UI(palette_inner_frame)]+1] end ]

    ac3d new_palette_entry $buttonid
}


set SCALEWIDTH 200



proc create_scaleXXX { name text entry} {
global SCALEWIDTH UI

    set len 5c
    set res 0.0039215686
    scale $name -width $SCALEWIDTH -label $text -from 0.0 -to 1.0 -length $len -orient horizontal -command "newcolour $entry" -resolution $res
}



proc get_palette_button_path { id} {
global UI

    set palframe $UI(palette_inner_frame)
    set palbuttonname [ format "$palframe.%04d" $id ]
	return $palbuttonname
}



proc newcolourXXX { index value } {
global prefs_realtimecolour
global progset
global current_palette_edit UI

    set index $current_palette_edit

#puts $index
        
        if { $progset == 0} {
          show_col_on_palbutton
          if { $prefs_realtimecolour} {
            ac3d set_palette_entry $index [.c.f1.diffuse_red get ] [.c.f1.diffuse_green get ]  [.c.f1.diffuse_blue get]\
                [.c.f2.ambient_red get ] [.c.f2.ambient_green get ]  [.c.f2.ambient_blue get] \
                [.c.f3.emissive_red get] [.c.f3.emissive_green get] [.c.f3.emissive_blue get]  \
                [.c.f4.specular_red get] [.c.f4.specular_green get] [.c.f4.specular_blue get] \
                [.c.f5.shininess get] [.c.f5.transparency get]
            ac3d redraw_3d
          }
        }

    } 

proc show_col_on_palbutton {} {

global current_palette_edit
global cp
global UI


    set palframe $UI(palette_inner_frame)

	set palbuttonname [get_palette_button_path $current_palette_edit]

#puts $palbuttonname

# perhaps average out diff/amb/emm for a better approximation for button????

    set r  [expr int([.c.f1.diffuse_red get ] * 255)]
    set g  [expr int([.c.f1.diffuse_green get ] * 255)]
    set b  [expr int([.c.f1.diffuse_blue get ] * 255)]
        
    set col  [format #%02x%02x%02x $r $g $b ]
        
    $palbuttonname configure  -background $col -activebackground $col

}

  
proc okpal { } {
global current_palette_edit current_palette_edit_name

    set index $current_palette_edit

	set mat [ac3d palette_get_material $index]
	set res [ ac3d entity_set_value $mat "name" $current_palette_edit_name]
}



# called when a coloured button e.g. diffuse is pressed - brings up system rgb col editor
proc material_edit_col { entry resname button} {
global prefs_realtimecolour

	set mat [ac3d palette_get_material $entry]
	set rgb [ ac3d entity_get_value $mat $resname]

	set deccol [rgbstring2dec $rgb]
	
	set color [tk_chooseColor -title "Material $entry $resname" -parent $button -initialcolor $deccol] 
#	puts "col $color"
	
	if {$color != "" } {
#		puts "set col on button and redraw views"
		
		if { $resname == "diffuse" } {
			set palbuttonname [get_palette_button_path $entry]
			$palbuttonname configure -background $color -activebackground $color
		}
		
		#set col on popup button
		$button configure -bg $color
		 
		set float3 [rgb2dec $color]
		scan $color "#%2x%2x%2x" r g b  
#		puts "RGB $r $g $b"
		set rf [expr $r/255.0]
		set gf [expr $g/255.0]
		set bf [expr $b/255.0]
#		puts "RGBF $rf $gf $bf"

	ac3d entity_set_value $mat $resname "$rf $gf $bf"
	palette_fill_values

	if { $prefs_realtimecolour } {
		ac3d redraw_all	
	}
	
	}
}



proc set_mat_from_rgb_slider { entry resourcename } {

	set mat [ac3d palette_get_material $entry]
	# update entity values from slider settings
	
	set r [$PALETTE($resourcename:rscale) get]
	set g [$PALETTE($resourcename:gscale) get]
	set b [$PALETTE($resourcename:bscale) get]

	set rgbstr "$r $g $b"
	ac3d entity_set_value $mat $resourcename $rgbstr
}



proc palette_slider_moved { resourcename sliderval} {
global PALETTE prefs_realtimecolour current_palette_entry

#puts "palette_slider_moved $resourcename $sliderval"

	set mat [ac3d palette_get_material $current_palette_entry]
	# update entity values from slider settings
	
	set r [$PALETTE($resourcename:rscale) get]
	set g [$PALETTE($resourcename:gscale) get]
	set b [$PALETTE($resourcename:bscale) get]

	# need to set these because they are being tracked by the scale widget and we need the value label to update
	set PALETTE($resourcename:r) $r
	set PALETTE($resourcename:g) $g
	set PALETTE($resourcename:b) $b

	set rgbstr "$r $g $b"
	ac3d entity_set_value $mat $resourcename $rgbstr
	
	# update button col
	set buttoncol [rgbstring2dec $rgbstr]
	$PALETTE($resourcename:button) configure -bg $buttoncol		
				
	if {$resourcename == "diffuse" } {
		set palbuttonname [get_palette_button_path $current_palette_entry]
		$palbuttonname configure -background $buttoncol -activebackground $buttoncol
	}
	
	if { $prefs_realtimecolour } {
		ac3d redraw_all	
	}
}


set palette_window_rows 1

# adds a single colour RGB entry e.g. emissive to the palette window

proc add_palette_window_col { entry resourcename labeltext top} {
global SCALEWIDTH PALETTE current_palette_entry

	set row [get_next_grid_row $top]

	ac_label $top.label$resourcename -text $labeltext

	set tf [ac_frame $top.frame$resourcename]

	# note use of old fashioned label here - because it's easy to mod the background col
	set bf [ac_frame $tf.bf -borderwidth 2 -relief sunken -style trough.TFrame]
	set PALETTE($resourcename:button) [::tk::label $bf.button$resourcename -text "       "  -borderwidth 0 -relief sunken ]
	bind $PALETTE($resourcename:button) <ButtonRelease-1> "material_edit_col \$current_palette_entry $resourcename $PALETTE($resourcename:button)"
	pack $PALETTE($resourcename:button)
	pack $bf -side left -anchor nw

	grid $top.label$resourcename -row $row -column 0 -sticky ne 
	grid $tf -row $row -column 1 -sticky nw 

	set f [ Fold:create $tf.fold$resourcename ]
	set cont [Fold:get_container $f]

	set PALETTE($resourcename:r) 0.1
	set PALETTE($resourcename:g) 0.2
	set PALETTE($resourcename:b) 0.999

	set sr [ac_create_pair_float_scale $cont PALETTE($resourcename:r) "R" "" "" 0.0 1.0 $SCALEWIDTH ]
	set sg [ac_create_pair_float_scale $cont PALETTE($resourcename:g) "G" "" "" 0.0 1.0 $SCALEWIDTH ]
	set sb [ac_create_pair_float_scale $cont PALETTE($resourcename:b) "B" "" "" 0.0 1.0 $SCALEWIDTH ]

	pack $f -side left

	set PALETTE($resourcename:rscale) $sr
	set PALETTE($resourcename:gscale) $sg
	set PALETTE($resourcename:bscale) $sb

	$PALETTE($resourcename:rscale) configure -command "palette_slider_moved $resourcename"
	$PALETTE($resourcename:gscale) configure -command "palette_slider_moved $resourcename"
	$PALETTE($resourcename:bscale) configure -command "palette_slider_moved $resourcename"
}


proc palette_fill_values {  } {
# read values from AC3D material and set the button cols and slider values
# uses current_palette_entry for the current entry index
global PALETTE current_palette_entry UI

	set mat [ac3d palette_get_material $current_palette_entry]
	set name [ ac3d entity_get_value $mat name]
	
	proc fill_resource { mat resourcename } {
	# fill for one named (rgbized) resource 
	global PALETTE
		set rgb [ ac3d entity_get_value $mat $resourcename]
		set buttoncol [rgbstring2dec $rgb]
		
		# set the col on the button
		$PALETTE($resourcename:button) configure -bg $buttoncol 		
		set	r [lindex $rgb 0]
		set g [lindex $rgb 1]
		set b [lindex $rgb 2]

		# set the values on the scale widgets
# these were calling the scale widget commands - so the col was being changed every time one value was set thus causing a redraw and flashing objects!
#		$PALETTE($resourcename:rscale) set $r
#		$PALETTE($resourcename:gscale) set $g
#		$PALETTE($resourcename:bscale) set $b

		set PALETTE($resourcename:r)  $r
		set PALETTE($resourcename:g)  $g
		set PALETTE($resourcename:b)  $b

	}
	
	fill_resource $mat "diffuse"
	fill_resource $mat "ambient"
	fill_resource $mat "emissive"
	fill_resource $mat "specular"
	
	set trans [ ac3d entity_get_value $mat "transparency"]
#	$PALETTE(transparency:scale) set $trans
	set PALETTE(transparency) $trans
		
	set shi [ ac3d entity_get_value $mat "shininess"]
#	$PALETTE(shininess:scale) set $shi	
	set PALETTE(shininess) $shi

	if { [info exists UI(material_data_text) ] == 1 } {
		if { [winfo exists $UI(material_data_text) ] } {
			set data [ ac3d entity_get_value $mat data]
			text_widget_set_text $UI(material_data_text)  $data
			}
	}
}


proc Fold:refresh { w } {
global FOLD

	set cs $FOLD($w:state)
	if { $cs == 1 } {
		pack $FOLD($w:subframe) -side left -expand 1 -fill both
	} else {
		pack forget $FOLD($w:subframe) 

	}

}

proc Fold:change_state { w } {
global FOLD

	if { $FOLD($w:state) == 0 } {
		set FOLD($w:state) 1
		$FOLD($w:button) configure -text " - "
	} else {
		set FOLD($w:state) 0
		$FOLD($w:button) configure -text " + "
	}
	
	Fold:refresh $w	
}


proc Fold:create { window { text "" } } {
global FOLD

	set f [ac_frame $window]
	set FOLD($window:state) 0
	set FOLD($window:topframe) [ac_frame $window.topf]
	pack $FOLD($window:topframe) -side top -fill x -expand 1
	
	set FOLD($window:button) [ac_button $window.topf.b -text " + " -command "Fold:change_state $window" -width 2]
	set FOLD($window:label) [ac_label $window.topf.label -text $text]
	pack $FOLD($window:button) -side left -expand 0
	pack $FOLD($window:label) -side left 
	pack $FOLD($window:topframe) -side top -fill x -expand 1

	set FOLD($window:subframe) [ac_frame $window.subframe]

	Fold:refresh $window
	return $window
}

proc Fold:get_container { w } {
global FOLD
	return $FOLD($w:subframe)
}

proc Fold:is_folded { w } {
global FOLD
	return !$FOLD($w:state)
}


proc palette_shininess_slider_moved { value } {
global PALETTE prefs_realtimecolour current_palette_entry

	set mat [ac3d palette_get_material $current_palette_entry]
	ac3d entity_set_value $mat shininess $value 

	if { $prefs_realtimecolour } {
		ac3d redraw_all	
	}
	
}


proc palette_transparency_slider_moved { value } {
global PALETTE prefs_realtimecolour current_palette_entry

	set mat [ac3d palette_get_material $current_palette_entry]
	ac3d entity_set_value $mat transparency $value 

	if { $prefs_realtimecolour } {
		ac3d redraw_all	
	}
}

proc palette_name_changed { } {
global PALETTE prefs_realtimecolour current_palette_entry current_palette_edit_name

	set mat [ac3d palette_get_material $current_palette_entry]
	ac3d entity_set_value $mat name $current_palette_edit_name
}


proc set_current_material_text { text } {
global PALETTE prefs_realtimecolour current_palette_entry current_palette_edit_name

	set mat [ac3d palette_get_material $current_palette_entry]
	ac3d entity_set_value $mat data $text
}



proc create_palette_window_content { parent } {
global UI prefs_realtimecolour SCALEWIDTH ac_platform
global prefs_allow_material_rename prefs_export_ac3d_version_c palette_window_rows
global current_palette_edit_name PALETTE current_palette_entry

	set entry $current_palette_entry


	set top [winfo toplevel $parent]
	wm title $top "Edit Material $entry"

    
	if { $prefs_allow_material_rename || $prefs_export_ac3d_version_c} {
		set ename [ ac_create_pair_label_entry $parent current_palette_edit_name "Name" ]
		bind  $ename <KeyRelease> { palette_name_changed }
	}



	add_palette_window_col $entry "diffuse" "Diffuse " $parent
	add_palette_window_col $entry "ambient" "Ambient " $parent
	add_palette_window_col $entry "emissive" "Emissive " $parent
	add_palette_window_col $entry "specular" "Specular " $parent


	global palette_shininess
	set PALETTE(shininess) 0
	set shiscale [ ac_create_pair_int_scale $parent PALETTE(shininess) "Shininess" "" palette_shininess_slider_moved 0 128 ]
	set PALETTE(shininess:scale) $shiscale
			
	global palette_trans
	set PALETTE(transparency) 0.0
	set transcale [ ac_create_pair_float_scale $parent PALETTE(transparency) "Transparency" "Transparency" palette_transparency_slider_moved 0.0 1.0 ]
	set PALETTE(transparency:scale) $transcale

	if { $prefs_export_ac3d_version_c } {
		set UI(material_data_text) [ ac_create_pair_label_text_expanding $parent "Data" "text which is saved along with the material" 40 3 ]

		bind $UI(material_data_text) <KeyRelease> {
			set_current_material_text  [$UI(material_data_text) get 0.0 end-1c]
			}
	}

	ac_create_pair_label_bool $parent prefs_realtimecolour "Realtime update " "Redraw the AC3D views everytime an update to the material is made?"

       
    palette_fill_values

}




proc refresh_palette_win { w } {
	palette_fill_values
}


proc create_palette_window { entry } {
global prefs_realtimecolour SCALEWIDTH ac_platform
global prefs_allow_material_rename palette_window_rows
global current_palette_edit_name PALETTE current_palette_entry

	set mat [ac3d palette_get_material $entry]
	set rgb [ ac3d entity_get_value $mat diffuse]
	set name [ ac3d entity_get_value $mat name]
	set current_palette_edit_name $name
	set current_palette_entry $entry
	
	ac_toplevel_with_close .matedit "Edit Material" create_palette_window_content refresh_palette_win

}





# rgb2dec --
#
#   Turns #rgb into 3 elem list of decimal vals.
#
# Arguments:
#   c		The #rgb hex of the color to translate
# Results:
#   Returns a #RRGGBB or #RRRRGGGGBBBB color
#
proc rgb2dec c {
    set c [string tolower $c]
    if {[regexp -nocase {^#([0-9a-f])([0-9a-f])([0-9a-f])$} $c x r g b]} {
	# double'ing the value make #9fc == #99ffcc
	scan "$r$r $g$g $b$b" "%x %x %x" r g b
    } else {
	if {![regexp {^#([0-9a-f]+)$} $c junk hex] || \
		[set len [string length $hex]]>12 || $len%3 != 0} {
	    if {[catch {winfo rgb . $c} rgb]} {
		return -code error "bad color value \"$c\""
	    } else {
		return $rgb
	    }
	}
	set len [expr {$len/3}]
    	scan $hex "%${len}x%${len}x%${len}x" r g b
    }
    return [list $r $g $b]
}

# dec2rgb --
#
#   Takes a color name or dec triplet and returns a #RRGGBB color.
#   If any of the incoming values are greater than 255,
#   then 16 bit value are assumed, and #RRRRGGGGBBBB is
#   returned, unless $clip is set.
#
# Arguments:
#   r		red dec value, or list of {r g b} dec value or color name
#   g		green dec value, or the clip value, if $r is a list
#   b		blue dec value
#   clip	Whether to force clipping to 2 char hex
# Results:
#   Returns a #RRGGBB or #RRRRGGGGBBBB color
#
proc dec2rgb {r {g 0} {b UNSET} {clip 0}} {
    if {![string compare $b "UNSET"]} {
	set clip $g
	if {[regexp {^-?(0-9)+$} $r]} {
	    foreach {r g b} $r {break}
	} else {
	    foreach {r g b} [winfo rgb . $r] {break}
	}
    } 
    set max 255
    set len 2
    if {($r > 255) || ($g > 255) || ($b > 255)} {
	if {$clip} {
	    set r [expr {$r>>8}]; set g [expr {$g>>8}]; set b [expr {$b>>8}]
	} else {
	    set max 65535
	    set len 4
	}
    }
    return [format "#%.${len}X%.${len}X%.${len}X" \
	    [expr {($r>$max)?$max:(($r<0)?0:$r)}] \
	    [expr {($g>$max)?$max:(($g<0)?0:$g)}] \
	    [expr {($b>$max)?$max:(($b<0)?0:$b)}]]
}

# shade --
#
#   Returns a shade between two colors
#
# Arguments:
#   orig	start #rgb color
#   dest	#rgb color to shade towards
#   frac	fraction (0.0-1.0) to move $orig towards $dest
# Results:
#   Returns a shade between two colors based on the
# 
proc shade {orig dest frac} {
    if {$frac >= 1.0} { return $dest } elseif {$frac <= 0.0} { return $orig }
    foreach {origR origG origB} [rgb2dec $orig] \
	    {destR destG destB} [rgb2dec $dest] {
	set shade [format "\#%02x%02x%02x" \
		[expr {int($origR+double($destR-$origR)*$frac)}] \
		[expr {int($origG+double($destG-$origG)*$frac)}] \
		[expr {int($origB+double($destB-$origB)*$frac)}]]
	return $shade
    }
}

# complement --
#
#   Returns a complementary color
#   Does some magic to avoid bad complements of grays
#
# Arguments:
#   orig	start #rgb color
# Results:
#   Returns a complement of a color
# 
proc complement {orig {grays 1}} {
    foreach {r g b} [rgb2dec $orig] {break}
    set R [expr {(~$r)%256}]
    set G [expr {(~$g)%256}]
    set B [expr {(~$b)%256}]
    if {$grays && abs($R-$r) < 32 && abs($G-$g) < 32 && abs($B-$b) < 32} {
	set R [expr {($r+128)%256}]
	set G [expr {($g+128)%256}]
	set B [expr {($b+128)%256}]
    }
    return [format "\#%02x%02x%02x" $R $G $B]
}



proc make_palette { w } {
	global UI

	set UI(palette_frame) $w
	set UI(palette_inner_frame) [ ScrolledHorizCanvas $UI(palette_frame) 10 22 {0 0 10 0} ]
	set UI(palette_canvas) $UI(palette_frame).canvas
	
	# this resizes the height of the outer canvas, according to the inner frame - e,g, when the font changes and when palette buttons are added
	bind $UI(palette_inner_frame) <Configure> {
		set win %W
		set h [ winfo height $win ]
		set w [ winfo width $win ]
		set rw [ winfo reqwidth $win ]
		$UI(palette_canvas) configure -height $h
		set region  "0 0 $rw $h"
		$UI(palette_canvas) configure -scrollregion $region
	}

	return $w
}















